# Copyright (c) 2019 NVIDIA Corporation
import os
import random
import time

import matplotlib
from matplotlib import pyplot as plt
import numpy as np
from sklearn.metrics import confusion_matrix, classification_report

from nemo.utils.exp_logging import get_logger
matplotlib.use("TkAgg")


logger = get_logger('')


def tensor2list(tensor):
    return tensor.detach().cpu().tolist()


def eval_iter_callback(tensors,
                       global_vars,
                       eval_data_layer):
    if "all_intent_preds" not in global_vars.keys():
        global_vars["all_intent_preds"] = []
    if "all_intent_labels" not in global_vars.keys():
        global_vars["all_intent_labels"] = []
    if "all_slot_preds" not in global_vars.keys():
        global_vars["all_slot_preds"] = []
    if "all_slot_labels" not in global_vars.keys():
        global_vars["all_slot_labels"] = []

    intent_logits_lists, intent_labels_lists = [], []
    slot_logits_lists, slot_labels_lists = [], []

    for kv, v in tensors.items():
        if kv.startswith('intent_logits'):
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    intent_logits_lists.append(tensor2list(logit_tensor))

        if kv.startswith('intents'):
            for v_tensor in v:
                for label_tensor in v_tensor:
                    intent_labels_lists.append(tensor2list(label_tensor))

        if kv.startswith('slot_logits'):
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    slot_logits_lists.append(tensor2list(logit_tensor))

        if kv.startswith('slots'):
            for v_tensor in v:
                for label_tensor in v_tensor:
                    slot_labels_lists.extend(tensor2list(label_tensor))

    intent_preds = list(np.argmax(np.asarray(intent_logits_lists), 1))
    slot_preds = list(np.argmax(np.asarray(slot_logits_lists), 2).flatten())
    global_vars["all_intent_preds"].extend(intent_preds)
    global_vars["all_intent_labels"].extend(intent_labels_lists)
    global_vars["all_slot_preds"].extend(slot_preds)
    global_vars["all_slot_labels"].extend(slot_labels_lists)


def list2str(l):
    return ' '.join([str(j) for j in l])


def eval_epochs_done_callback(global_vars, graph_fold):
    intent_labels = np.asarray(global_vars['all_intent_labels'])
    intent_preds = np.asarray(global_vars['all_intent_preds'])
    correct_preds = sum(intent_labels == intent_preds)
    intent_accuracy = correct_preds / intent_labels.shape[0]
    logger.info(f'Intent accuracy: {intent_accuracy}')

    slot_labels = np.asarray(global_vars['all_slot_labels'])
    slot_preds = np.asarray(global_vars['all_slot_preds'])
    slot_accuracy = sum(slot_labels == slot_preds) / slot_labels.shape[0]
    logger.info(f'Slot accuracy: {slot_accuracy}')

    i = 0
    if intent_preds.shape[0] > 21:
        i = random.randint(0, intent_preds.shape[0] - 21)
    logger.info("Sampled i_preds: [%s]" % list2str(intent_preds[i:i+20]))
    logger.info("Sampled intents: [%s]" % list2str(intent_labels[i:i+20]))
    logger.info("Sampled s_preds: [%s]" % list2str(slot_preds[i:i+20]))
    logger.info("Sampled slots: [%s]" % list2str(slot_labels[i:i+20]))
    cm = confusion_matrix(intent_labels, intent_preds)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    cax = ax.matshow(cm)
    plt.title('Confusion matrix of the classifier')
    fig.colorbar(cax)
    plt.xlabel('Predicted')
    plt.ylabel('True')
    os.makedirs(graph_fold, exist_ok=True)
    plt.savefig(os.path.join(graph_fold, time.strftime('%Y%m%d-%H%M%S')))

    logger.info(classification_report(intent_labels, intent_preds))

    return dict({'intent_accuracy': intent_accuracy,
                 'slot_accuracy': slot_accuracy})



# Copyright (c) 2019 NVIDIA Corporation


def eval_iter_callback(tensors, global_vars, eval_data_layer):
    if "correct_labels" not in global_vars.keys():
        global_vars["correct_labels"] = 0
    if "incorrect_labels" not in global_vars.keys():
        global_vars["incorrect_labels"] = 0
    if "correct_preds" not in global_vars.keys():
        global_vars["correct_preds"] = 0
    if "total_preds" not in global_vars.keys():
        global_vars["total_preds"] = 0
    if "total_correct" not in global_vars.keys():
        global_vars["total_correct"] = 0

    logits_lists = []
    seq_ids = []

    for kv, v in tensors.items():
        if 'logits' in kv:
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    logits_lists.append(logit_tensor.detach().cpu().tolist())

        if 'seq_ids' in kv:
            for v_tensor in v:
                for seq_id_tensor in v_tensor:
                    seq_ids.append(seq_id_tensor.detach().cpu().tolist())

    correct_labels, incorrect_labels, correct_preds, total_preds, \
        total_correct = eval_data_layer.eval_preds(logits_lists, seq_ids)

    global_vars["correct_labels"] += correct_labels
    global_vars["incorrect_labels"] += incorrect_labels
    global_vars["correct_preds"] += correct_preds
    global_vars["total_preds"] += total_preds
    global_vars["total_correct"] += total_correct


def eval_epochs_done_callback(global_vars):

    correct_labels = global_vars["correct_labels"]
    incorrect_labels = global_vars["incorrect_labels"]
    correct_preds = global_vars["correct_preds"]
    total_preds = global_vars["total_preds"]
    total_correct = global_vars["total_correct"]

    accuracy = correct_labels / (correct_labels + incorrect_labels)

    p = correct_preds / total_preds
    r = correct_preds / total_correct
    f1 = 2 * p * r / (p + r)

    print(f"Accuracy = {accuracy}")
    print(f"F1= {f1}")

    return dict({"accuracy": accuracy, "f1": f1})



from .bert_pretraining import BertPretrainingDataset
from .joint_intent_slot import BertJointIntentSlotDataset, \
    BertJointIntentSlotInferDataset
from .language_modeling import LanguageModelingDataset
from .ner import BertNERDataset
from .question_answering import BertQuestionAnsweringDataset
from .sentence_classification import BertSentenceClassificationDataset
from .token_classification import BertTokenClassificationDataset
from .translation import TranslationDataset



# =============================================================================
# Copyright 2019 AI Applications Design Team at NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""Pytorch Dataset for training BERT."""

import array
import glob
import os
import pickle
import random

import numpy as np
from torch.utils.data import Dataset
from tqdm import tqdm


class BertPretrainingDataset(Dataset):
    def __init__(self,
                 tokenizer,
                 dataset,
                 name,
                 sentence_indices_filename=None,
                 max_length=128,
                 mask_probability=0.15):
        self.tokenizer = tokenizer

        if sentence_indices_filename is None:
            sentence_indices_filename = "{}_sentence_indices.pkl".format(name)

        # Loading enormous datasets into RAM isn't always feasible -- for
        # example, the pubmed corpus is 200+ GB, which doesn't fit into RAM on
        # most computers. To get around this, we store the indices of newlines
        # in each file so we can seek to and retrieve sentences immediately
        # from main memory when needed during training.

        if os.path.isfile(sentence_indices_filename):
            # If the sentence indices file already exists, load from it
            with open(sentence_indices_filename, "rb") as f:
                sentence_indices = pickle.load(f)
        else:
            # Otherwise, generate and store sentence indices
            sentence_indices = {}
            total_tokens = 0
            used_tokens = 0

            # Finds all of the newline indices in a string
            def find_newlines(contents):
                nonlocal used_tokens, total_tokens
                start = 0

                while True:
                    try:
                        # index and split are much faster than Python for loops
                        new_start = contents.index(b"\n", start)
                        line = contents[start:new_start] \
                            .replace(b"\xc2\x99", b" ") \
                            .replace(b"\xc2\xa0", b" ") \
                            .replace(b"\xa0", b" ")
                        num_tokens = len(line.split())

                        # Ensure the line has at least max_length tokens
                        if num_tokens >= max_length:
                            yield start - 1
                            used_tokens += num_tokens

                        total_tokens += num_tokens
                        start = new_start + 1
                    except ValueError:
                        break

            if os.path.isdir(dataset):
                dataset_pattern = os.path.join(dataset, "**", "*.txt")
                filenames = glob.glob(dataset_pattern, recursive=True)
            else:
                filenames = [dataset]

            for filename in tqdm(filenames):
                with open(filename, "rb") as f:
                    contents = f.read()
                    newline_indices = find_newlines(contents)

                if os.path.isdir(dataset):
                    # Only keep the parts of the filepath that are invariant to
                    # the dataset's location on disk
                    filename = os.path.basename(filename)

                # In python, arrays are much more space-efficient than lists
                sentence_indices[filename] = array.array("I", newline_indices)

            # Save sentence indices so we don't have to do this again
            with open(sentence_indices_filename, "wb") as f:
                pickle.dump(sentence_indices, f)

            print("Used {} tokens of total {}".format(used_tokens,
                                                      total_tokens))

        corpus_size = 0
        empty_files = []

        # Find total number of newlines across entire corpus and remove files
        # without any newlines
        for filename in sentence_indices:
            if len(sentence_indices[filename]) <= 1:
                empty_files.append(filename)
            else:
                corpus_size += len(sentence_indices[filename])

        for filename in empty_files:
            del sentence_indices[filename]

        self.corpus_size = corpus_size
        self.dataset = dataset
        self.filenames = list(sentence_indices.keys())
        self.mask_probability = mask_probability
        self.max_length = max_length
        self.sentence_indices = sentence_indices
        self.vocab_size = self.tokenizer.vocab_size

    def __len__(self):
        return self.corpus_size

    def __getitem__(self, idx):
        # Each sequence has three special tokens, as follows:
        # [CLS] <document a> [SEP] <document b> [SEP]
        num_special_tokens = 3

        # TODO: Make seq_length = 512 for the last 10% of epochs, as specified
        # in BERT paper
        seq_length = self.max_length - num_special_tokens
        min_doc_length = 16

        a_length = random.randrange(min_doc_length,
                                    seq_length - min_doc_length + 1)
        b_length = seq_length - a_length

        a_filename = random.choice(self.filenames)
        a_line = random.choice(self.sentence_indices[a_filename])

        def get_document(filepath, line):
            # Retrieve a specific line from a file and return as a document
            if os.path.isdir(self.dataset):
                filepath = os.path.join(self.dataset, filepath)

            with open(filepath, "rb") as f:
                # Add one to go to the character after the newline
                f.seek(line + 1)

                # Read line, remove newline, and decode as UTF8
                doc_text = f.readline()[:-1].decode("utf-8", errors="ignore")
                document = [self.tokenizer.token_to_id("[CLS]")] \
                    + self.tokenizer.text_to_ids(doc_text) \
                    + [self.tokenizer.token_to_id("[SEP]")]

            assert len(document) >= self.max_length
            return document

        a_document = get_document(a_filename, a_line)

        if random.random() < 0.5:
            # 50% of the time, B is the sentence that follows A
            label = 1

            a_start_idx = random.randrange(len(a_document) - seq_length)
            b_start_idx = a_start_idx + a_length

            b_filename = a_filename
            b_line = a_line
            b_document = a_document
        else:
            # The rest of the time, B is a random sentence from the corpus
            label = 0

            a_start_idx = random.randrange(len(a_document) - a_length)

            b_filename = random.choice(self.filenames)
            b_line = random.choice(self.sentence_indices[b_filename])
            b_document = get_document(b_filename, b_line)
            b_start_idx = random.randrange(len(b_document) - b_length)

        # Process retrieved documents for use in training
        a_ids = a_document[a_start_idx:a_start_idx + a_length]
        b_ids = b_document[b_start_idx:b_start_idx + b_length]

        output_ids = [self.tokenizer.special_tokens["[CLS]"]] + a_ids + \
                     [self.tokenizer.special_tokens["[SEP]"]] + b_ids + \
                     [self.tokenizer.special_tokens["[SEP]"]]

        input_ids, output_mask = self.mask_ids(output_ids)

        output_mask = np.array(output_mask, dtype=np.float32)
        input_mask = np.ones(self.max_length, dtype=np.float32)

        input_type_ids = np.zeros(self.max_length, dtype=np.int)
        input_type_ids[a_length + 2:seq_length + 3] = 1

        return np.array(input_ids), input_type_ids, input_mask, \
            np.array(output_ids), output_mask, label

    def mask_ids(self, ids):
        """
        Args:
          tokens: list of tokens representing a chunk of text

        Returns:
          masked_tokens: list of input tokens with some of the entries masked
            according to the following protocol from the original BERT paper:
            each token is masked with a probability of 15% and is replaced with
            1) the [MASK] token 80% of the time,
            2) random token 10% of the time,
            3) the same token 10% of the time.
          output_mask: list of binary variables which indicate what tokens has
            been masked (to calculate the loss function for these tokens only)
        """
        masked_ids = []
        output_mask = []
        for id in ids:
            if random.random() < self.mask_probability:
                output_mask.append(1)
                if random.random() < 0.8:
                    masked_ids.append(self.tokenizer.special_tokens["[MASK]"])
                elif random.random() < 0.5:
                    masked_ids.append(random.randrange(self.vocab_size))
                else:
                    masked_ids.append(id)
            else:
                masked_ids.append(id)
                output_mask.append(0)
        return masked_ids, output_mask



# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions for Token Classification NLP tasks
Some transformer of this code were adapted from the HuggingFace library at
https://github.com/huggingface/pytorch-pretrained-BERT
"""

#TODO: REFACTOR to minimize code reusing


import collections
import numpy as np
from torch.utils.data import Dataset
import string
import re
import random
from ...externals.run_squad import _check_is_max_context


def remove_punctuation_from_sentence(sentence):
    sentence = re.sub('[' + string.punctuation + ']', '', sentence)
    sentence = sentence.lower()
    return sentence


class BertTokenClassificationDataset(Dataset):
    def __init__(self, input_file, max_seq_length, tokenizer):

        # Read the sentences and group them in sequences up to max_seq_length
        with open(input_file, "r") as f:
            self.seq_words = []
            self.seq_token_labels = []
            self.seq_sentence_labels = []
            self.seq_subtokens = []

            new_seq_words = []
            new_seq_token_labels = []
            new_seq_sentence_labels = []
            new_seq_subtokens = []
            new_seq_subtoken_count = 0

            lines = f.readlines()
            random.seed(0)
            random.shuffle(lines)

            for index, line in enumerate(lines):

                if index % 20000 == 0:
                    print(f"processing line {index}/{len(lines)}")

                sentence_label = line.split()[0]
                sentence = line.split()[2:]
                sentence = " ".join(sentence)
                # Remove punctuation
                sentence = remove_punctuation_from_sentence(sentence)
                sentence_words = sentence.split()

                sentence_subtoken_count = 0
                sentence_subtokens = []
                for word in sentence_words:
                    word_tokens = tokenizer.text_to_tokens(word)
                    sentence_subtokens.append(word_tokens)
                    sentence_subtoken_count += len(word_tokens)

                sentence_token_labels = [0] * sentence_subtoken_count
                sentence_token_labels[0] = 1

                # The -1 accounts for [CLS]
                max_tokens_for_doc = max_seq_length - 1

                if (new_seq_subtoken_count + sentence_subtoken_count) < \
                    max_tokens_for_doc:

                    new_seq_words.extend(sentence_words)
                    new_seq_token_labels.extend(sentence_token_labels)
                    new_seq_sentence_labels.append(sentence_label)
                    new_seq_subtokens.append(sentence_subtokens)
                    new_seq_subtoken_count += sentence_subtoken_count

                else:
                    self.seq_words.append(new_seq_words)
                    self.seq_token_labels.append(new_seq_token_labels)
                    self.seq_sentence_labels.append(new_seq_sentence_labels)
                    self.seq_subtokens.append(new_seq_subtokens)

                    new_seq_words = sentence_words
                    new_seq_token_labels = sentence_token_labels
                    new_seq_sentence_labels = [sentence_label]
                    new_seq_subtokens = [sentence_subtokens]
                    new_seq_subtoken_count = sentence_subtoken_count

        self.features = convert_sequences_to_features(
            self.seq_words, self.seq_subtokens, self.seq_token_labels,
            self.seq_sentence_labels, tokenizer, max_seq_length)

        self.tokenizer = tokenizer
        self.max_seq_length = max_seq_length
        self.vocab_size = self.tokenizer.vocab_size

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):

        feature = self.features[idx]

        return np.array(feature.input_ids), np.array(feature.segment_ids), \
            np.array(feature.input_mask, dtype=np.float32)[..., None], \
            np.array(feature.labels), np.array(feature.seq_id)

    def eval_preds(self, logits_lists, seq_ids):

        # Count the number of correct and incorrect predictions
        correct_labels = 0
        incorrect_labels = 0

        correct_preds = 0
        total_preds = 0
        total_correct = 0

        for logits, seq_id in zip(logits_lists, seq_ids):

            feature = self.features[seq_id]

            masks = feature.input_mask
            last_mask_index = masks.index(0)
            labels = feature.labels[:last_mask_index]
            labels = labels[:last_mask_index]
            logits = logits[:last_mask_index]

            preds = [1 if (a[1] > a[0]) else 0 for a in logits]

            correct_preds = 0
            correct_labels = 0
            for label, pred in zip(labels, preds):
                if pred == label:
                    correct_labels += 1
                    if pred == 1:
                        correct_preds += 1

            total_preds = preds.count(1)
            total_correct = labels.count(1)
            incorrect_labels = len(labels) - correct_labels

            if seq_id < 1:
                previous_word_id = -1
                predicted_seq = ""
                correct_seq = ""
                unpunctuated_seq = ""

                for token_id, word_id in feature.token_to_orig_map.items():

                    word = feature.words[word_id]

                    if word_id is not previous_word_id:
                        # New words has been found, handle it
                        if feature.labels[token_id] is 1:
                            if previous_word_id is not -1:
                                correct_seq += ". "
                            correct_seq += word.capitalize()
                        else:
                            correct_seq += " " + word

                        if preds[token_id] is 1:
                            if previous_word_id is not -1:
                                predicted_seq += ". "
                            predicted_seq += word.capitalize()
                        else:
                            predicted_seq += " " + word

                        unpunctuated_seq += " " + word

                    previous_word_id = word_id

                print("unpunctuated_seq:\n", unpunctuated_seq)
                print("correct_seq:\n", correct_seq)
                print("pred_seq:\n", predicted_seq)

        return correct_labels, incorrect_labels, correct_preds, total_preds, \
            total_correct


def convert_sequences_to_features(seqs_words, seqs_subtokens,
                                  seqs_token_labels, seqs_sentence_labels,
                                  tokenizer, max_seq_length):
    """Loads a data file into a list of `InputBatch`s."""

    features = []
    for seq_id, (words, seq_subtokens, seq_token_labels, sentence_labels) in \
        enumerate(zip(seqs_words, seqs_subtokens, seqs_token_labels,
            seqs_sentence_labels)):

        tok_to_orig_index = []
        orig_to_tok_index = []
        all_doc_tokens = []

        word_count = 0
        for sent_subtokens in seq_subtokens:
            for word_subtokens in sent_subtokens:
                orig_to_tok_index.append(len(all_doc_tokens))
                for sub_token in word_subtokens:
                    tok_to_orig_index.append(word_count)
                    all_doc_tokens.append(sub_token)
                word_count += 1

        _DocSpan = collections.namedtuple(  # pylint: disable=invalid-name
            "DocSpan", ["start", "length"])
        doc_spans = []
        start_offset = 0
        length = len(all_doc_tokens)
        doc_spans.append(_DocSpan(start=start_offset, length=length))

        doc_span_index = 0
        doc_span = doc_spans[0]

        tokens = []
        token_labels = []
        token_to_orig_map = {}
        token_is_max_context = {}
        segment_ids = []
        tokens.append("[CLS]")
        token_labels.append(0)
        segment_ids.append(0)

        for i in range(doc_span.length):
            split_token_index = doc_span.start + i
            token_to_orig_map[len(
                tokens)] = tok_to_orig_index[split_token_index]

            is_max_context = _check_is_max_context(doc_spans, doc_span_index,
                                                   split_token_index)
            token_is_max_context[len(tokens)] = is_max_context
            tokens.append(all_doc_tokens[split_token_index])
            segment_ids.append(0)

        for label in seq_token_labels:
            token_labels.append(label)

        input_ids = tokenizer.tokens_to_ids(tokens)

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        input_mask = [1] * len(input_ids)

        # Zero-pad up to the sequence length.
        while len(input_ids) < max_seq_length:
            input_ids.append(0)
            input_mask.append(0)
            segment_ids.append(0)
            token_labels.append(0)

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length
        assert len(token_labels) == max_seq_length

        if seq_id < 1:
            print("*** Example ***")
            print("example_index: %s" % seq_id)
            print("doc_span_index: %s" % doc_span_index)
            print("tokens: %s" % " ".join(tokens))
            print("words: %s" % " ".join(words))
            print("labels: %s" % " ".join(str(token_labels)))
            print("sentence_labels: %s" % " ".join(sentence_labels))
            print("token_to_orig_map: %s" % " ".join(
                ["%d:%d" % (x, y) for (x, y) in token_to_orig_map.items()]))
            print("token_is_max_context: %s" % " ".join(
                ["%d:%s" % (x, y) for (x, y) in token_is_max_context.items()]))
            print("input_ids: %s" % " ".join([str(x) for x in input_ids]))
            print("input_mask: %s" % " ".join([str(x) for x in input_mask]))
            print("segment_ids: %s" % " ".join([str(x) for x in segment_ids]))

        features.append(
            InputFeatures(seq_id=seq_id,
                          doc_span_index=doc_span_index,
                          tokens=tokens,
                          words=words,
                          labels=token_labels,
                          sentence_labels=sentence_labels,
                          token_to_orig_map=token_to_orig_map,
                          token_is_max_context=token_is_max_context,
                          input_ids=input_ids,
                          input_mask=input_mask,
                          segment_ids=segment_ids))

    return features


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self, seq_id, doc_span_index, tokens, words, labels,
                 sentence_labels, token_to_orig_map, token_is_max_context,
                 input_ids, input_mask, segment_ids):
        self.seq_id = seq_id
        self.doc_span_index = doc_span_index
        self.tokens = tokens
        self.words = words
        self.labels = labels
        self.sentence_labels = sentence_labels
        self.token_to_orig_map = token_to_orig_map
        self.token_is_max_context = token_is_max_context
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids



from .tokenizer_spec import TokenizerSpec
from pytorch_transformers import BertTokenizer
import re


def handle_quotes(text):
    text_ = ""
    quote = 0
    i = 0
    while i < len(text):
        if text[i] == "\"":
            if quote % 2:
                text_ = text_[:-1] + "\""
            else:
                text_ += "\""
                i += 1
            quote += 1
        else:
            text_ += text[i]
        i += 1
    return text_


def remove_spaces(text):
    text = text.replace("( ", "(")
    text = text.replace(" )", ")")
    text = text.replace("[ ", "[")
    text = text.replace(" ]", "]")
    text = text.replace(" / ", "/")
    text = text.replace("„ ", "„")
    text = text.replace(" - ", "-")
    text = text.replace(" ' ", "'")
    text = re.sub(r'([0-9])( )([\.,])', '\\1\\3', text)
    text = re.sub(r'([\.,])( )([0-9])', '\\1\\3', text)
    text = re.sub(r'([0-9])(:)( )([0-9])', '\\1\\2\\4', text)
    text = text.replace(" %", "%")
    text = text.replace("$ ", "$")
    text = text.replace("\xa0", " ")
    text = re.sub(r'([^0-9])(,)([0-9])', '\\1\\2 \\3', text)
    return text


class NemoBertTokenizer(TokenizerSpec):
    def __init__(self, pretrained_model=None,
                 vocab_file=None,
                 do_lower_case=True,
                 max_len=None,
                 do_basic_tokenize=True,
                 never_split=("[UNK]", "[SEP]", "[PAD]", "[CLS]", "[MASK]")):
        if pretrained_model:
            self.tokenizer = BertTokenizer.from_pretrained(pretrained_model)
            if "uncased" not in pretrained_model:
                self.tokenizer.basic_tokenizer.do_lower_case = False
        else:
            self.tokenizer = BertTokenizer(vocab_file,
                                           do_lower_case,
                                           max_len,
                                           do_basic_tokenize,
                                           never_split)
        self.vocab_size = len(self.tokenizer.vocab)
        self.never_split = never_split

    def text_to_tokens(self, text):
        tokens = self.tokenizer.tokenize(text)
        return tokens

    def tokens_to_text(self, tokens):
        text = self.tokenizer.convert_tokens_to_string(tokens)
        return remove_spaces(handle_quotes(text.strip()))

    def token_to_id(self, token):
        return self.tokens_to_ids([token])[0]

    def tokens_to_ids(self, tokens):
        ids = self.tokenizer.convert_tokens_to_ids(tokens)
        return ids

    def ids_to_tokens(self, ids):
        tokens = self.tokenizer.convert_ids_to_tokens(ids)
        return tokens

    def text_to_ids(self, text):
        tokens = self.text_to_tokens(text)
        ids = self.tokens_to_ids(tokens)
        return ids

    def ids_to_text(self, ids):
        tokens = self.ids_to_tokens(ids)
        tokens_clean = [t for t in tokens if t not in self.never_split]
        text = self.tokens_to_text(tokens_clean)
        return text

    def pad_id(self):
        return self.tokens_to_ids(["[PAD]"])[0]

    def bos_id(self):
        return self.tokens_to_ids(["[CLS]"])[0]

    def eos_id(self):
        return self.tokens_to_ids(["[SEP]"])[0]



import sentencepiece as spm
from .tokenizer_spec import TokenizerSpec


class SentencePieceTokenizer(TokenizerSpec):
    def __init__(self, model_path):
        self.tokenizer = spm.SentencePieceProcessor()
        self.tokenizer.Load(model_path)
        self.original_vocab_size = self.tokenizer.get_piece_size()
        self.vocab_size = self.tokenizer.get_piece_size()
        self.special_tokens = {}
        self.special_token_ids = {}

    def text_to_tokens(self, text):
        tokens = []
        idx = 0
        last_idx = 0

        while 1:
            indices = {}

            for token in self.special_tokens:
                try:
                    indices[token] = text[idx:].index(token)
                except ValueError:
                    continue

            if len(indices) == 0:
                break

            next_token = min(indices, key=indices.get)
            next_idx = idx + indices[next_token]

            tokens.extend(self.tokenizer.encode_as_pieces(text[idx:next_idx]))
            tokens.append(next_token)
            idx = next_idx + len(next_token)

        tokens.extend(self.tokenizer.encode_as_pieces(text[idx:]))
        return tokens

    def tokens_to_text(self, tokens):
        return self.tokenizer.decode_pieces(tokens)

    def text_to_ids(self, text):
        ids = []
        idx = 0
        last_idx = 0

        while 1:
            indices = {}

            for token in self.special_tokens:
                try:
                    indices[token] = text[idx:].index(token)
                except ValueError:
                    continue

            if len(indices) == 0:
                break

            next_token = min(indices, key=indices.get)
            next_idx = idx + indices[next_token]

            ids.extend(self.tokenizer.encode_as_ids(text[idx:next_idx]))
            ids.append(self.special_tokens[next_token])
            idx = next_idx + len(next_token)

        ids.extend(self.tokenizer.encode_as_ids(text[idx:]))
        return ids

    def ids_to_text(self, ids):
        text = ""
        last_i = 0

        for i, id in enumerate(ids):
            if id in self.special_token_ids:
                text += self.tokenizer.decode_ids(ids[last_i:i]) + " "
                text += self.special_token_ids[id] + " "
                last_i = i + 1

        text += self.tokenizer.decode_ids(ids[last_i:])
        return text.strip()

    def token_to_id(self, token):
        if token in self.special_tokens:
            return self.special_tokens[token]

        return self.tokenizer.piece_to_id(token)

    def tokens_to_ids(self, tokens):
        ids = []

        for token in tokens:
            if token in self.special_tokens:
                ids.append(self.special_tokens[token])
            else:
                ids.append(self.tokenizer.piece_to_id(token))

        return ids

    def ids_to_tokens(self, ids):
        tokens = []

        for id in ids:
            if id >= self.original_vocab_size:
                tokens.append(self.special_token_ids[id])
            else:
                tokens.append(self.tokenizer.id_to_piece(id))

        return tokens

    def add_special_tokens(self, special_tokens):
        for token in special_tokens:
            if self.tokenizer.piece_to_id(token) == self.tokenizer.unk_id():
                self.special_tokens[token] = self.vocab_size
                self.special_token_ids[self.vocab_size] = token
                self.vocab_size += 1



# Copyright (c) 2019 NVIDIA Corporation
"""
This package contains BERT Neural Module
"""
import torch
import torch.nn as nn

from nemo.backends.pytorch.nm import TrainableNM, LossNM
from nemo.core.neural_types import *
from .transformer import ClassificationLogSoftmax
from .transformer import SmoothedCrossEntropyLoss
from .transformer import SequenceClassificationLoss
from .transformer.utils import transformer_weights_init


class MaskedLanguageModelingLossNM(TrainableNM):
    @staticmethod
    def create_ports():
        input_ports = {
            "log_probs":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            }),
            "output_ids":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "output_mask":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
        }

        output_ports = {"loss": NeuralType(None)}
        return input_ports, output_ports

    def __init__(self, **kwargs):
        TrainableNM.__init__(self, **kwargs)
        label_smoothing = self.local_parameters.get("label_smoothing", 0.0)
        self._loss_fn = SmoothedCrossEntropyLoss(label_smoothing)

    def forward(self, log_probs, output_ids, output_mask):
        loss = self._loss_fn(log_probs, output_ids, output_mask)
        return loss


class SentenceClassificationLogSoftmaxNM(TrainableNM):
    @staticmethod
    def create_ports():
        input_ports = {
            "hidden_states":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            }),
        }

        output_ports = {
            "log_probs":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            }),
        }
        return input_ports, output_ports

    def __init__(self, *, d_model, num_classes, **kwargs):
        TrainableNM.__init__(self, **kwargs)

        self.log_softmax = ClassificationLogSoftmax(
            hidden_size=d_model,
            num_classes=num_classes
        )

        self.log_softmax.apply(transformer_weights_init)
        self.log_softmax.to(self._device)

    def forward(self, hidden_states):
        log_probs = self.log_softmax(hidden_states)
        return log_probs


class NextSentencePredictionLossNM(TrainableNM):
    @staticmethod
    def create_ports():
        input_ports = {
            "log_probs":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            }),
            "labels":
            NeuralType({0: AxisType(BatchTag)}),
        }

        output_ports = {"loss": NeuralType(None)}
        return input_ports, output_ports

    def __init__(self, **kwargs):
        TrainableNM.__init__(self, **kwargs)
        self._loss_fn = SequenceClassificationLoss()

    def forward(self, log_probs, labels):
        loss = self._loss_fn(log_probs, labels)
        return loss


class LossAggregatorNM(LossNM):
    @staticmethod
    def create_ports(num_losses=2):
        input_ports = {}
        for i in range(num_losses):
            input_ports["loss_" + str(i + 1)] = NeuralType(None)

        output_ports = {"loss": NeuralType(None)}
        return input_ports, output_ports

    def __init__(self, *, num_inputs, **kwargs):
        kwargs["create_port_args"] = {"num_losses": num_inputs}
        LossNM.__init__(self, **kwargs)

    def _loss_function(self, **kwargs):
        values = [kwargs[x] for x in sorted(kwargs.keys())]
        loss = values[0]
        for loss_i in values[1:]:
            loss = loss.add(loss_i.item())
        return loss


class QuestionAnsweringPredictionLoss(TrainableNM):
    @staticmethod
    def create_ports():
        input_ports = {
            "hidden_states":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            }),
            "start_positions":
            NeuralType({0: AxisType(BatchTag)}),
            "end_positions":
            NeuralType({0: AxisType(BatchTag)})
        }

        output_ports = {
            "loss":
            NeuralType(None),
            "start_logits":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "end_logits":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            })
        }
        return input_ports, output_ports

    def __init__(self, **kwargs):
        TrainableNM.__init__(self, **kwargs)

        self.hidden_size = self.local_parameters["d_model"]

        self.qa_outputs = nn.Linear(self.hidden_size, 2)
        self.qa_outputs.apply(transformer_weights_init)
        self.qa_outputs.to(self._device)

    def forward(self, hidden_states, start_positions, end_positions):

        logits = self.qa_outputs(hidden_states)

        start_logits, end_logits = logits.split(1, dim=-1)
        start_logits = start_logits.squeeze(-1)
        end_logits = end_logits.squeeze(-1)

        # If we are on multi-GPU, split add a dimension
        if len(start_positions.size()) > 1:
            start_positions = start_positions.squeeze(-1)
        if len(end_positions.size()) > 1:
            end_positions = end_positions.squeeze(-1)

        # sometimes the start/end positions are outside our model inputs,
        # we ignore these terms
        ignored_index = start_logits.size(1)
        start_positions.clamp_(0, ignored_index)
        end_positions.clamp_(0, ignored_index)

        loss_fct = nn.CrossEntropyLoss(ignore_index=ignored_index)
        start_loss = loss_fct(start_logits, start_positions)
        end_loss = loss_fct(end_logits, end_positions)
        total_loss = (start_loss + end_loss) / 2

        return total_loss, start_logits, end_logits


class TokenClassificationLoss(TrainableNM):
    @staticmethod
    def create_ports():
        input_ports = {
            "hidden_states":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            }),
            "labels":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_mask":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            })
        }

        output_ports = {
            "loss":
            NeuralType(None),
            "logits":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            })
        }
        return input_ports, output_ports

    def __init__(self, **kwargs):
        TrainableNM.__init__(self, **kwargs)

        self.hidden_size = self.local_parameters["d_model"]
        self.num_labels = self.local_parameters["num_labels"]
        self.dropout = nn.Dropout(self.local_parameters["dropout"])
        self.classifier = nn.Linear(self.hidden_size, self.num_labels)

        self.apply(
            lambda module: transformer_weights_init(module, xavier=False))
        self.to(self._device)

    def forward(self, hidden_states, labels, input_mask):

        hidden_states = self.dropout(hidden_states)
        logits = self.classifier(hidden_states)

        loss_fct = nn.CrossEntropyLoss()

        active_loss = input_mask.view(-1) > 0.5
        active_logits = logits.view(-1, self.num_labels)[active_loss]
        active_labels = labels.view(-1)[active_loss]

        loss = loss_fct(active_logits, active_labels)

        return loss, logits


class ZerosLikeNM(TrainableNM):
    @staticmethod
    def create_ports():
        input_ports = {
            "input_type_ids":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
            })
        }

        output_ports = {
            "input_type_ids":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
            })
        }
        return input_ports, output_ports

    def __init__(self, **kwargs):
        TrainableNM.__init__(self, **kwargs)

    def forward(self, input_type_ids):
        return torch.zeros_like(input_type_ids).long()


class SequenceClassifier(TrainableNM):
    """
    Loss function for the joint intent classification and slot
    filling task.

    The loss is a joint loss of both tasks, aim to maximize:
    p(y^i | x)P(y^s1, y^s2, ..., y^sn | x)

    with y^i being the predicted intent and y^s1, y^s2, ..., y^sn
    are the predicted slots corresponding to x1, x2, ..., xn.

    Args:
        hidden_states: output of the hidden layers
        intents: ground truth intents,
        slots: ground truth slots.
        input_mask: to differentiate from original tokens and paddings
        intent_loss_weight: the loss is the sum of:
            intent_loss_weight * intent_loss +
            (1 - intent_loss_weight) * slot_loss

    """
    @staticmethod
    def create_ports():
        input_ports = {
            "hidden_states": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            })
        }

        output_ports = {
            "logits": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(ChannelTag)
            }),
        }
        return input_ports, output_ports

    def __init__(self, hidden_size, num_classes, dropout, **kwargs):
        TrainableNM.__init__(self, **kwargs)
        self.hidden_size = hidden_size
        self.num_classes = num_classes
        self.dropout = nn.Dropout(dropout)
        self.dense = nn.Linear(self.hidden_size, self.hidden_size)
        self.classifier = nn.Linear(self.hidden_size, self.num_classes)
        self.apply(
            lambda module: transformer_weights_init(module, xavier=False))
        self.to(self._device)

    def forward(self, hidden_states):
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.dense(hidden_states[:, 0])
        hidden_states = torch.relu(hidden_states)
        logits = self.classifier(hidden_states)

        return logits


class JointIntentSlotClassifier(TrainableNM):
    """
    Loss function for the joint intent classification and slot
    filling task.

    The loss is a joint loss of both tasks, aim to maximize:
    p(y^i | x)P(y^s1, y^s2, ..., y^sn | x)

    with y^i being the predicted intent and y^s1, y^s2, ..., y^sn
    are the predicted slots corresponding to x1, x2, ..., xn.

    Args:
        hidden_states: output of the hidden layers
        intents: ground truth intents,
        slots: ground truth slots.
        input_mask: to differentiate from original tokens and paddings
        intent_loss_weight: the loss is the sum of:
            intent_loss_weight * intent_loss +
            (1 - intent_loss_weight) * slot_loss

    """
    @staticmethod
    def create_ports():
        input_ports = {
            "hidden_states": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            }),
            "input_mask": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
        }

        output_ports = {
            "intent_logits": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(ChannelTag)
            }),
            "slot_logits": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            })
        }
        return input_ports, output_ports

    def __init__(self,
                 hidden_size,
                 num_intents,
                 num_slots,
                 dropout,
                 **kwargs):
        TrainableNM.__init__(self, **kwargs)
        self.hidden_size = hidden_size
        self.num_intents = num_intents
        self.num_slots = num_slots
        self.dropout = nn.Dropout(dropout)
        self.intent_dense = nn.Linear(self.hidden_size, self.hidden_size)
        self.intent_classifier = nn.Linear(self.hidden_size, self.num_intents)
        self.slot_dense = nn.Linear(self.hidden_size, self.hidden_size)
        self.slot_classifier = nn.Linear(self.hidden_size, self.num_slots)
        self.apply(
            lambda module: transformer_weights_init(module, xavier=False))
        self.to(self._device)

    def forward(self, hidden_states):
        hidden_states = self.dropout(hidden_states)

        intent_states = self.intent_dense(hidden_states[:, 0])
        intent_states = torch.relu(intent_states)
        intent_logits = self.intent_classifier(intent_states)

        # slot_states = self.slot_dense(hidden_states[1:, :])
        slot_states = self.slot_dense(hidden_states)
        slot_states = torch.relu(slot_states)
        slot_logits = self.slot_classifier(slot_states)

        return intent_logits, slot_logits


class JointIntentSlotLoss(LossNM):
    """
    Loss function for the joint intent classification and slot
    filling task.

    The loss is a joint loss of both tasks, aim to maximize:
    p(y^i | x)P(y^s1, y^s2, ..., y^sn | x)

    with y^i being the predicted intent and y^s1, y^s2, ..., y^sn
    are the predicted slots corresponding to x1, x2, ..., xn.

    Args:
        hidden_states: output of the hidden layers
        intents: ground truth intents,
        slots: ground truth slots.
        input_mask: to differentiate from original tokens and paddings
        intent_loss_weight: the loss is the sum of:
            intent_loss_weight * intent_loss +
            (1 - intent_loss_weight) * slot_loss

    """
    @staticmethod
    def create_ports():
        input_ports = {
            "intent_logits": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(ChannelTag)
            }),
            "slot_logits": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(ChannelTag)
            }),
            "input_mask": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "intents": NeuralType({
                0: AxisType(BatchTag),
            }),
            "slots":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            # "intent_loss_weight": NeuralType(None)
        }

        output_ports = {
            "loss": NeuralType(None),
        }
        return input_ports, output_ports

    def __init__(self, num_slots, **kwargs):
        LossNM.__init__(self, **kwargs)
        self.num_slots = num_slots
        self._criterion = nn.CrossEntropyLoss()

    def _loss_function(self,
                       intent_logits,
                       slot_logits,
                       input_mask,
                       intents,
                       slots,
                       intent_loss_weight=0.6):
        intent_loss = self._criterion(intent_logits, intents)

        active_loss = input_mask.view(-1) > 0.5
        active_logits = slot_logits.view(-1, self.num_slots)[active_loss]
        active_labels = slots.view(-1)[active_loss]

        slot_loss = self._criterion(active_logits, active_labels)
        loss = intent_loss * intent_loss_weight + \
            slot_loss * (1 - intent_loss_weight)

        return loss



# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for NER NLP tasks
Some transformer of this code were adapted from the HuggingFace library at
https://github.com/huggingface/pytorch-pretrained-BERT
"""

# TODO: REFACTOR to minimize code reusing

import collections
import numpy as np
from torch.utils.data import Dataset


class BertNERDataset(Dataset):
    def __init__(self, input_file, max_seq_length, tokenizer):
        # Read the sentences and group them in sequences up to max_seq_length
        with open(input_file, "r") as f:
            self.seq_words = []
            self.seq_token_labels = []
            self.seq_subtokens = []

            new_seq_words = []
            new_seq_token_labels = []
            new_seq_subtokens = []
            new_seq_subtoken_count = 0

            lines = f.readlines()

            words = []
            tags = []
            tokens = []
            token_tags = []
            token_count = 0

            def process_sentence():
                nonlocal new_seq_words, new_seq_token_labels, \
                    new_seq_subtokens, new_seq_subtoken_count

                # The -1 accounts for [CLS]
                max_tokens_for_doc = max_seq_length - 1

                if (new_seq_subtoken_count + token_count) < max_tokens_for_doc:
                    new_seq_words.extend(words)
                    new_seq_token_labels.extend(token_tags)
                    new_seq_subtokens.append(tokens)
                    new_seq_subtoken_count += token_count
                else:
                    self.seq_words.append(new_seq_words)
                    self.seq_token_labels.append(new_seq_token_labels)
                    self.seq_subtokens.append(new_seq_subtokens)

                    new_seq_words = words
                    new_seq_token_labels = token_tags
                    new_seq_subtokens = [tokens]
                    new_seq_subtoken_count = token_count

            all_tags = {}

            # Collect a list of all possible tags
            for line in lines:
                if line == "\n":
                    continue

                tag = line.split()[-1]

                if tag not in tags and tag != "O":
                    tag = tag.split("-")[1]
                    all_tags[tag] = 0

            # Create mapping of tags to tag ids that starts with "O" -> 0 and
            # then increases in alphabetical order
            tag_ids = {"O": 0}

            for tag in sorted(all_tags):
                tag_ids[tag] = len(tag_ids)

            # Process all lines in input data
            for line in lines:
                if line == "\n":
                    # If we hit a newline, we've reached the end of a sentence
                    process_sentence()
                    words = []
                    tags = []
                    tokens = []
                    token_tags = []
                    continue

                word = line.split()[0]
                tag = line.split()[-1]

                if tag != "O":
                    tag = tag.split("-")[1]

                word_tokens = tokenizer.text_to_tokens(word)
                tag_id = tag_ids[tag]

                words.append(word)
                tags.append(tag_id)
                tokens.append(word_tokens)
                token_tags.extend([tag_id] * len(word_tokens))
                token_count += len(word_tokens)

            self.features = convert_sequences_to_features(
                self.seq_words, self.seq_subtokens, self.seq_token_labels,
                tokenizer, max_seq_length)

            self.tag_ids = tag_ids
            self.tokenizer = tokenizer
            self.max_seq_length = max_seq_length
            self.vocab_size = self.tokenizer.vocab_size

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):
        feature = self.features[idx]

        return np.array(feature.input_ids), np.array(feature.segment_ids), \
            np.array(feature.input_mask, dtype=np.float32), \
            np.array(feature.labels), np.array(feature.seq_id)

    def eval_preds(self, logits_lists, seq_ids, tag_ids):
        correct_chunks = 0
        total_chunks = 0
        predicted_chunks = 0

        correct_tags = 0
        token_count = 0

        lines = []
        tag_ids = {tag_ids[k]: k for k in tag_ids}

        for logits, seq_id in zip(logits_lists, seq_ids):
            feature = self.features[seq_id]
            masks = feature.input_mask

            try:
                last_mask_index = masks.index(0)
            except ValueError:
                last_mask_index = len(masks)

            labels = feature.labels[:last_mask_index]
            labels = np.array(labels[:last_mask_index])
            logits = logits[:last_mask_index]
            preds = np.argmax(logits, axis=1)

            in_correct = False

            last_label = "O"
            last_pred = "O"

            last_correct = "O"
            last_correct_type = ""
            last_guessed = "O"
            last_guessed_type = ""

            previous_word_id = -1

            # Adapted from conlleval.pl (included with CoNLL-2003 dataset)
            for token_id, word_id in feature.token_to_orig_map.items():
                if word_id is not previous_word_id:
                    word = feature.words[word_id]
                    label = tag_ids[feature.labels[token_id]]
                    pred = tag_ids[preds[token_id]]

                    if pred != "O":
                        guessed = "B" if last_pred != pred else "I"
                        guessed_type = pred
                    else:
                        guessed = pred
                        guessed_type = ""

                    if label != "O":
                        correct = "B" if last_label != label else "I"
                        correct_type = label
                    else:
                        correct = label
                        correct_type = ""

                    if in_correct:
                        if end_of_chunk(last_correct, correct,
                                        last_correct_type, correct_type) and \
                                end_of_chunk(last_guessed, guessed,
                                             last_guessed_type,
                                             guessed_type) and \
                                last_guessed_type == last_correct_type:
                            in_correct = False
                            correct_chunks += 1

                        elif end_of_chunk(last_correct, correct,
                                          last_correct_type, correct_type) != \
                                end_of_chunk(last_guessed, guessed,
                                             last_guessed_type,
                                             guessed_type) or \
                                guessed_type != correct_type:
                            in_correct = False

                    if start_of_chunk(last_correct, correct, last_correct_type,
                                      correct_type) and \
                            start_of_chunk(last_guessed, guessed,
                                           last_guessed_type,
                                           guessed_type) and \
                            guessed_type == correct_type:
                        in_correct = True

                    if start_of_chunk(last_correct, correct, last_correct_type,
                                      correct_type):
                        total_chunks += 1

                    if start_of_chunk(last_guessed, guessed, last_guessed_type,
                                      guessed_type):
                        predicted_chunks += 1

                    if correct == guessed and guessed_type == correct_type:
                        correct_tags += 1

                    token_count += 1

                    last_label = label
                    last_pred = pred

                    last_guessed = guessed
                    last_correct = correct
                    last_guessed_type = guessed_type
                    last_correct_type = correct_type

                    lines.append({
                        "word": word,
                        "label": feature.labels[token_id],
                        "prediction": preds[token_id]
                    })

                previous_word_id = word_id

            if in_correct:
                correct_chunks += 1

            lines.append({
                "word": ""
            })

        return correct_tags, token_count, correct_chunks, predicted_chunks, \
            total_chunks, lines


def start_of_chunk(prev_tag, tag, prev_type, type):
    if prev_tag == "B" and tag == "B":
        return True
    elif prev_tag == "I" and tag == "B":
        return True
    elif prev_tag == "O" and tag == "B":
        return True
    elif prev_tag == "O" and tag == "I":
        return True
    elif prev_tag == "O" and tag == "I":
        return True
    elif tag != "O" and prev_type != type:
        return True

    return False


def end_of_chunk(prev_tag, tag, prev_type, type):
    if prev_tag == "B" and tag == "B":
        return True
    elif prev_tag == "B" and tag == "O":
        return True
    elif prev_tag == "I" and tag == "B":
        return True
    elif prev_tag == "I" and tag == "O":
        return True
    elif prev_tag == "I" and tag == "O":
        return True
    elif prev_tag != "O" and prev_type != type:
        return True

    return False


def convert_sequences_to_features(seqs_words, seqs_subtokens,
                                  seqs_token_labels, tokenizer,
                                  max_seq_length):
    """Loads a data file into a list of `InputBatch`s."""

    features = []
    for seq_id, (words, seq_subtokens, seq_token_labels) in \
            enumerate(zip(seqs_words, seqs_subtokens, seqs_token_labels)):

        tok_to_orig_index = []
        orig_to_tok_index = []
        all_doc_tokens = []

        word_count = 0
        for sent_subtokens in seq_subtokens:
            for word_subtokens in sent_subtokens:
                orig_to_tok_index.append(len(all_doc_tokens))
                for sub_token in word_subtokens:
                    tok_to_orig_index.append(word_count)
                    all_doc_tokens.append(sub_token)
                word_count += 1

        _DocSpan = collections.namedtuple(  # pylint: disable=invalid-name
            "DocSpan", ["start", "length"])
        doc_spans = []
        start_offset = 0
        length = len(all_doc_tokens)
        doc_spans.append(_DocSpan(start=start_offset, length=length))

        doc_span_index = 0
        doc_span = doc_spans[0]

        tokens = []
        token_labels = []
        token_to_orig_map = {}
        token_is_max_context = {}
        segment_ids = []
        tokens.append("[CLS]")
        token_labels.append(0)
        segment_ids.append(0)

        # Ensure that we don't go over the maximum sequence length
        for i in range(min(doc_span.length, max_seq_length - 1)):
            split_token_index = doc_span.start + i
            token_to_orig_map[len(tokens)] = \
                tok_to_orig_index[split_token_index]

            is_max_context = _check_is_max_context(doc_spans, doc_span_index,
                                                   split_token_index)
            token_is_max_context[len(tokens)] = is_max_context
            tokens.append(all_doc_tokens[split_token_index])
            segment_ids.append(0)

        for label in seq_token_labels:
            if len(token_labels) == len(tokens):
                break

            token_labels.append(label)

        input_ids = tokenizer.tokens_to_ids(tokens)

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        input_mask = [1] * len(input_ids)

        # Zero-pad up to the sequence length.
        while len(input_ids) < max_seq_length:
            input_ids.append(0)
            input_mask.append(0)
            segment_ids.append(0)
            token_labels.append(0)

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length
        assert len(token_labels) == max_seq_length

        features.append(
            InputFeatures(
                seq_id=seq_id,
                doc_span_index=doc_span_index,
                tokens=tokens,
                words=words,
                labels=token_labels,
                token_to_orig_map=token_to_orig_map,
                token_is_max_context=token_is_max_context,
                input_ids=input_ids,
                input_mask=input_mask,
                segment_ids=segment_ids))

    return features


def _check_is_max_context(doc_spans, cur_span_index, position):
    """Check if this is the 'max context' doc span for the token."""

    # Because of the sliding window approach taken to scoring documents, a
    # single token can appear in multiple documents. E.g.
    #  Doc: the man went to the store and bought a gallon of milk
    #  Span A: the man went to the
    #  Span B: to the store and bought
    #  Span C: and bought a gallon of
    #  ...
    #
    # Now the word 'bought' will have two scores from spans B and C. We only
    # want to consider the score with "maximum context", which we define as
    # the *minimum* of its left and right context (the *sum* of left and
    # right context will always be the same, of course).
    #
    # In the example the maximum context for 'bought' would be span C since
    # it has 1 left context and 3 right context, while span B has 4 left
    # context and 0 right context.
    best_score = None
    best_span_index = None
    for (span_index, doc_span) in enumerate(doc_spans):
        end = doc_span.start + doc_span.length - 1
        if position < doc_span.start:
            continue
        if position > end:
            continue
        num_left_context = position - doc_span.start
        num_right_context = end - position
        score = min(num_left_context, num_right_context) + \
            0.01 * doc_span.length
        if best_score is None or score > best_score:
            best_score = score
            best_span_index = span_index

    return cur_span_index == best_span_index


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self,
                 seq_id,
                 doc_span_index,
                 tokens,
                 words,
                 labels,
                 token_to_orig_map,
                 token_is_max_context,
                 input_ids,
                 input_mask,
                 segment_ids):
        self.seq_id = seq_id
        self.doc_span_index = doc_span_index
        self.tokens = tokens
        self.words = words
        self.labels = labels
        self.token_to_orig_map = token_to_orig_map
        self.token_is_max_context = token_is_max_context
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids



# Copyright (c) 2019 NVIDIA Corporation
import numpy as np

from nemo.utils.exp_logging import get_logger

GLOBAL_KEYS = ["eval_loss", "sys"]

logger = get_logger('')


def eval_iter_callback(tensors, global_vars):

    for key in GLOBAL_KEYS:
        if key not in global_vars.keys():
            global_vars[key] = []

    for kv, v in tensors.items():
        if "loss" in kv:
            for eval_loss in v:
                global_vars["eval_loss"].append(eval_loss.item())


def eval_epochs_done_callback(global_vars):
    eval_loss = np.mean(global_vars["eval_loss"])
    eval_ppl = np.exp(eval_loss)

    logger.info("------------------------------------------------------------")
    logger.info("Eval loss: {0}".format(np.round(eval_loss, 3)))
    logger.info("Eval  ppl: {0}".format(np.round(eval_ppl, 3)))
    logger.info("------------------------------------------------------------")
    for key in GLOBAL_KEYS:
        global_vars[key] = []
    return dict({"Eval_loss": eval_loss, "Eval_ppl": eval_ppl})



import torch
from torch import nn


class SmoothedCrossEntropyLoss(nn.Module):
    """
    Cross-entropy loss with label smoothing for a batch of sequences.

    Args:
        label_smoothing: label smoothing coefficient, usually set between 0.0
            and 0.1 in language modeling and translation pipelines
        predict_last_k: int parameter which sets the number of last tokens to
            calculate the loss for, for example
            0: (default) calculate loss on the entire sequence (e.g., NMT)
            1: calculate loss on the last token only (e.g., LM evaluation)
            Intermediate values allow to control the trade-off between eval
            time (proportional to the number of batches) and eval performance
            (proportional to the number of context tokens).
    """

    def __init__(self, label_smoothing=0.0, predict_last_k=0):
        super().__init__()
        self._smoothing = label_smoothing
        self._predict_last_k = predict_last_k

    def forward(self, log_probs, output_ids, output_mask):
        """
        Args:
            log_probs: float tensor of shape batch_size x seq_len x vocab_size
            output_ids: int tensor of shape batch_size x seq_len
            output_mask: binary tensor of shape batch_size x seq_len
        """
        batch_size, seq_len, vocab_size = log_probs.size()
        smoothing = vocab_size * self._smoothing / (vocab_size - 1)
        target_log_probs = log_probs.gather(
            2, output_ids.unsqueeze(2)).squeeze(2)
        smoothing_log_probs = log_probs.mean(dim=-1)
        neg_log_likelihood = (1.0 - smoothing) * target_log_probs + \
            smoothing * smoothing_log_probs
        neg_log_likelihood = neg_log_likelihood[:, -self._predict_last_k:]
        output_mask = output_mask[:, -self._predict_last_k:]
        neg_log_likelihood = -torch.sum(neg_log_likelihood * output_mask)
        neg_log_likelihood = neg_log_likelihood / (output_mask.sum() + 1e-6)
        return neg_log_likelihood


class SequenceClassificationLoss(nn.Module):
    """
    Sequence classification loss.
    """

    def __init__(self):
        super().__init__()

    def forward(self, log_probs, labels):
        """
        Args:
            log_probs: float tensor of shape batch_size x num_classes
            labels: int tensor of shape batch_size
        """
        log_probs_target = log_probs.gather(1, labels.unsqueeze(1))
        neg_log_likelihood = -torch.mean(log_probs_target)
        return neg_log_likelihood



# Copyright 2019 AI Applications Design Team at NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from .data import *
from .transformer_nm import TransformerEncoderNM, TransformerDecoderNM, \
    TransformerLogSoftmaxNM, PaddedSmoothedCrossEntropyLossNM, \
    BeamSearchTranslatorNM, GreedyLanguageGeneratorNM
from .bert import MaskedLanguageModelingLossNM, \
    SentenceClassificationLogSoftmaxNM, NextSentencePredictionLossNM, \
    LossAggregatorNM, QuestionAnsweringPredictionLoss, \
    TokenClassificationLoss, SequenceClassifier, \
    JointIntentSlotLoss, ZerosLikeNM, \
    JointIntentSlotClassifier
from .nlp_utils import read_intent_slot_outputs
from . import transformer, huggingface

from .callbacks import *

from nemo.core import Backend


name = "nemo_nlp"
backend = Backend.PyTorch
__version__ = "0.1"



from .bert_ner_data_layer import BertNERDataLayer
from .bert_pretraining_data_layer import BertPretrainingDataLayer
from .bert_qa_data_layer import BertQuestionAnsweringDataLayer
from .bert_tc_data_layer import BertTokenClassificationDataLayer
from .datasets import *
from .text_data_layers import *
from .tokenizers import *
from .translation_data_layer import TranslationDataLayer



# Copyright (c) 2019 NVIDIA Corporation

from nemo.backends.pytorch.nm import DataLayerNM
from nemo.core.neural_types import *
from nemo.core import DeviceType
import torch
from .datasets import BertTokenClassificationDataset


class BertTokenClassificationDataLayer(DataLayerNM):
    @staticmethod
    def create_ports():
        input_ports = {}
        output_ports = {
            "input_ids":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_type_ids":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_mask":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag),
                2: AxisType(TimeTag)
            }),
            "labels":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "seq_ids":
            NeuralType({0: AxisType(BatchTag)})
        }

        return input_ports, output_ports

    def __init__(self, *, tokenizer, path_to_data, max_seq_length, **kwargs):
        DataLayerNM.__init__(self, **kwargs)
        self._dataset = BertTokenClassificationDataset(
            tokenizer=tokenizer,
            input_file=path_to_data,
            max_seq_length=max_seq_length)

    def eval_preds(self, logits, seq_ids):
        correct_labels, incorrect_labels, correct_preds, total_preds, \
            total_correct = self._dataset.eval_preds(logits, seq_ids)
        return correct_labels, incorrect_labels, correct_preds, total_preds, \
            total_correct

    def __len__(self):
        return len(self._dataset)

    @property
    def dataset(self):
        return self._dataset

    @property
    def data_iterator(self):
        return None



# Copyright (c) 2019 NVIDIA Corporation

from nemo.backends.pytorch.nm import DataLayerNM
from nemo.core.neural_types import *
from nemo.core import DeviceType
import torch
from .datasets import BertPretrainingDataset


class BertPretrainingDataLayer(DataLayerNM):
    @staticmethod
    def create_ports():
        input_ports = {}
        output_ports = {
            "input_ids":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_type_ids":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_mask":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "output_ids":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "output_mask":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "labels":
            NeuralType({0: AxisType(BatchTag)}),
        }

        return input_ports, output_ports

    def __init__(self, *, tokenizer, dataset, name, max_seq_length,
                 sentence_indices_filename=None, mask_probability=0.15,
                 **kwargs):
        DataLayerNM.__init__(self, **kwargs)

        self._device = torch.device(
            "cuda" if self.placement in [DeviceType.GPU, DeviceType.AllGpu]
            else "cpu"
        )

        self._dataset = BertPretrainingDataset(
            tokenizer=tokenizer,
            dataset=dataset,
            name=name,
            sentence_indices_filename=sentence_indices_filename,
            max_length=max_seq_length,
            mask_probability=mask_probability)

    def __len__(self):
        return len(self._dataset)

    @property
    def dataset(self):
        return self._dataset

    @property
    def data_iterator(self):
        return None



# Copyright (c) 2019 NVIDIA Corporation
import numpy as np


def eval_iter_callback(tensors, global_vars):
    if "dev_mlm_loss" not in global_vars.keys():
        global_vars["dev_mlm_loss"] = []
    if "dev_nsp_loss" not in global_vars.keys():
        global_vars["dev_nsp_loss"] = []
    keys = list(tensors.keys())
    # TODO: referring to these by name here is error-prone
    for dev_mlm_loss in tensors[keys[1]]:
        global_vars["dev_mlm_loss"].append(dev_mlm_loss.item())

    if len(keys) > 2:
        for dev_nsp_loss in tensors[keys[2]]:
            global_vars["dev_nsp_loss"].append(dev_nsp_loss.item())


def eval_epochs_done_callback(global_vars):
    if 'dev_mlm_loss' in global_vars:
        mlm_loss = np.mean(global_vars["dev_mlm_loss"])
        print("Dev MLM perplexity: {0}".format(np.round(np.exp(mlm_loss), 3)))
        global_vars["dev_mlm_loss"] = []
    else:
        mlm_loss = -123.0

    if 'dev_nsp_loss' in global_vars:
        nsp_loss = np.mean(global_vars["dev_nsp_loss"])
        print("Dev NSP perplexity: {0}".format(np.round(np.exp(nsp_loss), 3)))
        global_vars["dev_nsp_loss"] = []
    else:
        nsp_loss = -123.0

    return dict({"Dev MLM loss": mlm_loss, "Dev NSP loss": nsp_loss})



# Copyright 2018 The Google AI Language Team Authors and
# The HuggingFace Inc. team.
# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for Token Classification NLP tasks
Some parts of this code were adapted from the HuggingFace library at
https://github.com/huggingface/pytorch-pretrained-BERT
"""

import collections
import logging
import os
import random
import string
import time

import numpy as np
from torch.utils.data import Dataset

logger = logging.getLogger('log')


def get_stats(lengths):
    lengths = np.asarray(lengths)
    logger.info(f'Min: {np.min(lengths)} | \
                 Max: {np.max(lengths)} | \
                 Mean: {np.mean(lengths)} | \
                 Median: {np.median(lengths)}')
    print(f'75 percentile: {np.percentile(lengths, 75)} | \
            99 percentile: {np.percentile(lengths, 99)}')


def list2str(l):
    return ' '.join([str(x) for x in l])


class BertSentenceClassificationDataset(Dataset):
    """A dataset class that converts from raw data to
    a dataset that can be used by DataLayerNM.

    Args:
        input_file (str): file to sequence + label.
            the first line is header (sentence [tab] label)
            each line should be [sentence][tab][label]
        max_seq_length (int): max sequence length minus 2 for [CLS] and [SEP]
        tokenizer (Tokenizer): such as BertTokenizer
        num_samples (int): number of samples you want to use for the dataset.
            If -1, use all dataset. Useful for testing.
        shuffle (bool): whether to shuffle your data.
    """

    def __init__(self,
                 input_file,
                 max_seq_length,
                 tokenizer,
                 num_samples=-1,
                 shuffle=True):
        with open(input_file, "r") as f:
            sent_labels, all_sent_subtokens = [], []
            sent_lengths = []
            too_long_count = 0

            lines = f.readlines()[1:]
            logger.info(f'{input_file}: {len(lines)}')

            if shuffle or num_samples > -1:
                random.seed(0)
                random.shuffle(lines)
                if num_samples > 0:
                    lines = lines[:num_samples]

            for index, line in enumerate(lines):
                if index % 20000 == 0:
                    logger.debug(f"Processing line {index}/{len(lines)}")

                sent_label = int(line.split()[-1])
                sent_labels.append(sent_label)
                sent_words = line.strip().split()[:-1]
                sent_subtokens = ['[CLS]']

                for word in sent_words:
                    word_tokens = tokenizer.tokenize(word)
                    sent_subtokens.extend(word_tokens)

                sent_subtokens.append('[SEP]')

                all_sent_subtokens.append(sent_subtokens)
                sent_lengths.append(len(sent_subtokens))

        get_stats(sent_lengths)
        self.max_seq_length = min(max_seq_length, max(sent_lengths))

        for i in range(len(all_sent_subtokens)):
            if len(all_sent_subtokens[i]) > self.max_seq_length:
                shorten_sent = all_sent_subtokens[i][-self.max_seq_length+1:]
                all_sent_subtokens[i] = ['[CLS]'] + shorten_sent
                too_long_count += 1

        logger.info(f'{too_long_count} out of {len(sent_lengths)} \
                       sentencess with more than {max_seq_length} subtokens.')

        self.convert_sequences_to_features(all_sent_subtokens,
                                           sent_labels,
                                           tokenizer,
                                           self.max_seq_length)

        self.tokenizer = tokenizer
        self.vocab_size = self.tokenizer.vocab_size

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):

        feature = self.features[idx]

        return (np.array(feature.input_ids),
                np.array(feature.segment_ids),
                np.array(feature.input_mask, dtype=np.float32),
                feature.sent_label)

    def convert_sequences_to_features(self,
                                      all_sent_subtokens,
                                      sent_labels,
                                      tokenizer,
                                      max_seq_length):
        """Loads a data file into a list of `InputBatch`s.
        """

        self.features = []
        for sent_id in range(len(all_sent_subtokens)):
            sent_subtokens = all_sent_subtokens[sent_id]
            sent_label = sent_labels[sent_id]
            word_count = 0
            # input_ids = tokenizer.tokens_to_ids(sent_subtokens)
            input_ids = [tokenizer._convert_token_to_id(
                t) for t in sent_subtokens]

            # The mask has 1 for real tokens and 0 for padding tokens.
            # Only real tokens are attended to.
            input_mask = [1] * len(input_ids)

            # Zero-pad up to the sequence length.
            while len(input_ids) < max_seq_length:
                input_ids.append(0)
                input_mask.append(0)
            segment_ids = [0] * max_seq_length

            assert len(input_ids) == max_seq_length
            assert len(input_mask) == max_seq_length

            if sent_id == 0:
                logger.info("*** Example ***")
                logger.info("example_index: %s" % sent_id)
                logger.info("subtokens: %s" % " ".join(sent_subtokens))
                logger.info("sent_label: %s" % sent_label)
                logger.info("input_ids: %s" % list2str(input_ids))
                logger.info("input_mask: %s" % list2str(input_mask))

            self.features.append(InputFeatures(
                sent_id=sent_id,
                sent_label=sent_label,
                input_ids=input_ids,
                input_mask=input_mask,
                segment_ids=segment_ids))


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self,
                 sent_id,
                 sent_label,
                 input_ids,
                 input_mask,
                 segment_ids):
        self.sent_id = sent_id
        self.sent_label = sent_label
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids



# Copyright (c) 2019 NVIDIA Corporation
import os
import random
import time

import logging

import matplotlib
from matplotlib import pyplot as plt
import numpy as np
from sklearn.metrics import confusion_matrix, classification_report

matplotlib.use("TkAgg")
logger = logging.getLogger('log')


def eval_iter_callback(tensors,
                       global_vars,
                       eval_data_layer):
    if "all_preds" not in global_vars.keys():
        global_vars["all_preds"] = []
    if "all_labels" not in global_vars.keys():
        global_vars["all_labels"] = []

    logits_lists = []
    labels_lists = []

    for kv, v in tensors.items():
        if 'logits' in kv:
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    logits_lists.append(logit_tensor.detach().cpu().tolist())

        if 'labels' in kv:
            for v_tensor in v:
                for label_tensor in v_tensor:
                    labels_lists.append(label_tensor.detach().cpu().tolist())

    preds = list(np.argmax(np.asarray(logits_lists), 1))
    global_vars["all_preds"].extend(preds)
    global_vars["all_labels"].extend(labels_lists)


def list2str(l):
    return ' '.join([str(j) for j in l])


def eval_epochs_done_callback(global_vars, graph_fold):
    labels = np.asarray(global_vars['all_labels'])
    preds = np.asarray(global_vars['all_preds'])
    accuracy = sum(labels == preds) / labels.shape[0]
    logger.info(f'Accuracy: {accuracy}')
    i = 0
    if preds.shape[0] > 21:
        i = random.randint(0, preds.shape[0] - 21)
    logger.info("Sampled preds: [%s]" % list2str(preds[i:i+20]))
    logger.info("Sampled labels: [%s]" % list2str(labels[i:i+20]))
    cm = confusion_matrix(labels, preds)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    cax = ax.matshow(cm)
    plt.title('Confusion matrix of the classifier')
    fig.colorbar(cax)
    plt.xlabel('Predicted')
    plt.ylabel('True')
    os.makedirs(graph_fold, exist_ok=True)
    plt.savefig(os.path.join(graph_fold, time.strftime('%Y%m%d-%H%M%S')))

    logger.info(classification_report(labels, preds))

    return dict({"accuracy": accuracy})



# Copyright (c) 2019 NVIDIA Corporation


def eval_iter_callback(tensors, global_vars):
    if "eval_start_logits" not in global_vars.keys():
        global_vars["eval_start_logits"] = []
    if "eval_end_logits" not in global_vars.keys():
        global_vars["eval_end_logits"] = []
    if "eval_unique_ids" not in global_vars.keys():
        global_vars["eval_unique_ids"] = []

    for kv, v in tensors.items():

        if 'logits' in kv:
            logits = []
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    logits.append(logit_tensor.detach().cpu().tolist())

            if kv.startswith('start_logits'):
                global_vars['eval_start_logits'].extend(logits)
            elif kv.startswith('end_logits'):
                global_vars['eval_end_logits'].extend(logits)

        if 'unique_ids' in kv:
            unique_ids = []
            for v_tensor in v:
                for id_tensor in v_tensor:
                    unique_ids.append(id_tensor.detach().cpu().tolist())
            global_vars['eval_unique_ids'].extend(unique_ids)


def eval_epochs_done_callback(global_vars, eval_data_layer, do_lower_case):

    exact_match, f1 = eval_data_layer.calculate_exact_match_and_f1(
        global_vars["eval_unique_ids"],
        global_vars["eval_start_logits"],
        global_vars["eval_end_logits"],
        do_lower_case=do_lower_case)

    print(f"Exact_match = {exact_match}, f1 = {f1}")

    global_vars["eval_unique_ids"] = []
    global_vars["eval_start_logits"] = []
    global_vars["eval_end_logits"] = []

    return dict({"exact_match": exact_match, "f1": f1})



# Copyright (c) 2019 NVIDIA Corporation
"""
If you want to add your own data layer, you should put its name in
__all__ so that it can be imported with 'from text_data_layers import *'
"""
__all__ = ['TextDataLayer',
           'BertSentenceClassificationDataLayer',
           'BertJointIntentSlotDataLayer',
           'BertJointIntentSlotInferDataLayer',
           'LanguageModelingDataLayer']

import torch

import nemo
from nemo.backends.pytorch.nm import DataLayerNM
from nemo.core.neural_types import *


class TextDataLayer(DataLayerNM):
    """
    Generic Text Data Layer NM which wraps PyTorch's dataset


    Args:
        dataset: a PyTorch dataset to wrap into Neural Module

    """
    def __init__(self, dataset, **kwargs):
        DataLayerNM.__init__(self, **kwargs)
        self._dataset = dataset

    def __len__(self):
        return len(self._dataset)

    @property
    def dataset(self):
        return self._dataset

    @property
    def data_iterator(self):
        return None


class BertSentenceClassificationDataLayer(TextDataLayer):
    """
    Creates the data layer to use for the task of sentence classification
    with pretrained model.

    All the data processing is done BertSentenceClassificationDataset.

    Args:
        dataset (BertSentenceClassificationDataset):
                the dataset that needs to be converted to DataLayerNM
    """

    @staticmethod
    def create_ports():
        output_ports = {
            "input_ids": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_type_ids": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_mask": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "labels": NeuralType({
                0: AxisType(BatchTag),
            }),
        }
        return {}, output_ports


class BertJointIntentSlotDataLayer(TextDataLayer):
    """
    Creates the data layer to use for the task of joint intent
    and slot classification with pretrained model.

    All the data processing is done in BertJointIntentSlotDataset.

    Args:
        dataset (BertJointIntentSlotDataset):
                the dataset that needs to be converted to DataLayerNM
    """
    @staticmethod
    def create_ports():
        output_ports = {
            "input_ids": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_type_ids": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_mask": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "token_mask": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "intents": NeuralType({
                0: AxisType(BatchTag),
            }),
            "slots": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
        }
        return {}, output_ports


class BertJointIntentSlotInferDataLayer(TextDataLayer):
    """
    Creates the data layer to use for the task of joint intent
    and slot classification with pretrained model. This is for

    All the data processing is done in BertJointIntentSlotInferDataset.

    Args:
        dataset (BertJointIntentSlotInferDataset):
                the dataset that needs to be converted to DataLayerNM
    """
    @staticmethod
    def create_ports():
        output_ports = {
            "input_ids": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_type_ids": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_mask": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "token_mask": NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            })
        }
        return {}, output_ports


class LanguageModelingDataLayer(TextDataLayer):
    @staticmethod
    def create_ports():
        input_ports = {}
        output_ports = {
            "input_ids":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "input_mask":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            }),
            "labels":
            NeuralType({
                0: AxisType(BatchTag),
                1: AxisType(TimeTag)
            })
        }

        return input_ports, output_ports



# Copyright (c) 2019 NVIDIA Corporation


def eval_iter_callback(tensors, global_vars, eval_data_layer, tag_ids):
    if "correct_tags" not in global_vars.keys():
        global_vars["correct_tags"] = 0
    if "token_count" not in global_vars.keys():
        global_vars["token_count"] = 0
    if "correct_chunks" not in global_vars.keys():
        global_vars["correct_chunks"] = 0
    if "predicted_chunks" not in global_vars.keys():
        global_vars["predicted_chunks"] = 0
    if "total_chunks" not in global_vars.keys():
        global_vars["total_chunks"] = 0
    if "lines" not in global_vars.keys():
        global_vars["lines"] = []

    logits_lists = []
    seq_ids = []

    for kv, v in tensors.items():
        if "logits" in kv:
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    logits_lists.append(logit_tensor.detach().cpu().tolist())

        if "seq_ids" in kv:
            for v_tensor in v:
                for seq_id_tensor in v_tensor:
                    seq_ids.append(seq_id_tensor.detach().cpu().tolist())

    correct_tags, token_count, correct_chunks, predicted_chunks, \
        total_chunks, lines = \
        eval_data_layer.eval_preds(logits_lists, seq_ids, tag_ids)

    global_vars["correct_tags"] += correct_tags
    global_vars["token_count"] += token_count
    global_vars["correct_chunks"] += correct_chunks
    global_vars["predicted_chunks"] += predicted_chunks
    global_vars["total_chunks"] += total_chunks
    global_vars["lines"].extend(lines)


def eval_epochs_done_callback(global_vars, tag_ids, output_filename):
    correct_tags = global_vars["correct_tags"]
    token_count = global_vars["token_count"]
    correct_chunks = global_vars["correct_chunks"]
    predicted_chunks = global_vars["predicted_chunks"]
    total_chunks = global_vars["total_chunks"]
    lines = global_vars["lines"]

    if output_filename is not None:
        # Create output file that can be evaluated by conlleval.pl script
        tag_ids = {tag_ids[k]: k for k in tag_ids}

        last_label = ""
        last_prediction = ""

        with open(output_filename, "w") as f:
            for line in lines:
                if line["word"] == "":
                    f.write("\n")
                    last_label = ""
                    last_prediction = ""
                    continue

                label = tag_ids[int(line["label"])]
                prediction = tag_ids[int(line["prediction"])]

                # Correctly precede tags with B- and I- as necessary (slightly
                # modified from https://www.clips.uantwerpen.be/conll2003/ner/)
                if label != "O":
                    if last_label == line["label"]:
                        label = "I-{}".format(label)
                    else:
                        label = "B-{}".format(label)

                if prediction != "O":
                    if last_prediction == line["prediction"]:
                        prediction = "I-{}".format(prediction)
                    else:
                        prediction = "B-{}".format(prediction)

                last_label = line["label"]
                last_prediction = line["prediction"]

                f.write("{}\t{}\t{}\n".format(line["word"], label, prediction))

    accuracy = correct_tags / token_count

    p = correct_chunks / predicted_chunks if predicted_chunks > 0 else 0
    r = correct_chunks / total_chunks if total_chunks > 0 else 0
    f1 = 2 * p * r / (p + r) if p > 0 and r > 0 else 0

    print(f"Accuracy = {accuracy}")
    print(f"F1 = {f1}")

    return {"accuracy": accuracy, "f1": f1}



from .spc_tokenizer import SentencePieceTokenizer
from .bert_tokenizer import NemoBertTokenizer
from .yttm_tokenizer import YouTokenToMeTokenizer
from .gpt2_tokenizer import NemoGPT2Tokenizer
from .word_tokenizer import WordTokenizer
from .char_tokenizer import CharTokenizer



from abc import abstractmethod, ABC
from typing import List


class TokenizerSpec(ABC):
    @abstractmethod
    def text_to_tokens(self, text):
        pass

    @abstractmethod
    def tokens_to_text(self, tokens):
        pass

    @abstractmethod
    def tokens_to_ids(self, tokens):
        pass

    @abstractmethod
    def ids_to_tokens(self, ids):
        pass

    @abstractmethod
    def text_to_ids(self, text):
        pass

    @abstractmethod
    def ids_to_text(self, ids):
        pass

    def add_special_tokens(self, special_tokens: List[str]):
        pass
